#!/usr/bin/python3

import lldb
import sys
import os
import re
import shlex
import argparse


# Global constants
# This path assumes that the script is stored in Reveal's Application Support directory under 'Scripts'.
applicationSupportDirectory = os.path.expanduser("~/Library/Application Support/Reveal/")
localFrameworkPath = os.path.join(applicationSupportDirectory, 'RevealServer', 'RevealServer.xcframework')

# Entry point
def __lldb_init_module(debugger, internal_dict):
    # Make the options parser so we can generate the help text for the LLDB command prior to registering it below
    HandleRevealCommand.__doc__ = CreateRevealCommandOptionsParser().format_help()
    # Install the debugger command
    debugger.HandleCommand("command script add -f {0}.HandleRevealCommand reveal".format(__name__))

# Command options parsing
def CreateRevealCommandOptionsParser():
    subcommands = {}
    subcommands["load"] = { "func" : HandleRevealLoadCommand, "help" : "Load the Reveal Server library." }
    subcommands["start"] = { "func" : HandleRevealStartCommand, "help" : "Start Reveal Server, if it's loaded." }
    subcommands["stop"] = { "func" : HandleRevealStopCommand, "help" : "Stop Reveal Server." }
    subcommands["status"] = { "func" : HandleRevealStatusCommand, "help" : "Print the current status of Reveal Server." }
    
    description = "Commands for loading and controlling Reveal Server in the current debugging session."
    parser = argparse.ArgumentParser(prog="reveal", description=description)
    subparsers = parser.add_subparsers(title="Available actions", metavar="action")
    
    subparsersMap = {}
    for key, info in subcommands.items():
        subparser = subparsers.add_parser(key, help=info["help"])
        subparser.set_defaults(func=info["func"])
        subparsersMap[key] = subparser
    
    subparsersMap["load"].add_argument("-a", "--autostart", action="store_true", help="Automatically start Reveal Server after loading.")
    
    return parser

# Command handlers
def HandleRevealCommand(debugger, command, exe_ctx, result, internal_dict):
    # Parse the command
    parser = CreateRevealCommandOptionsParser()
    commandArgs = shlex.split(command)
    
    if len(commandArgs) == 0:
        parser.print_help()
        return
    
    try:
        args = parser.parse_args(commandArgs)
    except:
        return
    
    # Bail out if running an incompatible target
    target = exe_ctx.target
    if TargetIsCompatible(target) == False:
        result.SetError("target {0} is not supported by Reveal Server.".format(target.triple))
        return
    
    # Check that process is already in the right state
    process = exe_ctx.process
    if lldb.SBDebugger.StateIsStoppedState(process.state) == False:
        result.SetError("process must be paused to execute Reveal Server commands.")
        return

    if ProcessMainThreadContainsFrameWithName(process, "UIApplicationMain") == False:
        if ProcessMainThreadContainsFrameWithName(process, "NSExtensionMain") == False:
            result.SetError("process is not yet ready to execute Reveal Server commands.")
        else:
            result.SetError("process is running an unsupported application extension. Cannot load Reveal Server.")
        return
    
    # Create a loader object with main thread's current frame and execute the command with it
    mainThreadFrame = process.GetThreadAtIndex(0).GetSelectedFrame()
    loader = RevealLoader(mainThreadFrame)
    args.func(loader, result, args)

def HandleRevealLoadCommand(loader, result, args):
    if IsLegacyBreakpointAtProcessFirstFrame(loader.process):
        SetupLegacyLoadBreakpoints(loader.process.target, args.autostart, result)
        return

    # If the target is running in the Simulator, load Reveal Server from the local host
    if TargetIsSimulator(loader.process.target):
        binaryPath = loader.localRevealServerBinaryPath(loader.process.target, result)

        if binaryPath is not None:
            loader.injectServer(binaryPath, result)

    else:
        # If target is running on device, check if Reveal Server framework is included in the bundle and try loading it if it exists
        remoteServerBinaryPath = loader.remoteRevealServerBinaryPath(loader.process.target, result)

        if remoteServerBinaryPath is not None and loader.remoteFileExists(remoteServerBinaryPath):
            loader.injectServer(remoteServerBinaryPath, result)
        else:
            # Otherwise, can't load Reveal
            result.SetError("failed to load Reveal Server because it was not found in the application bundle at {0}. For information about debugging apps with Reveal on device, please refer to the Integration Guide.".format(remoteServerBinaryPath))
            return
    
    # If injection succeeded and autostart option is specified, start the Server
    if result.Succeeded() and args.autostart == True:
        loader.startServer()

def HandleRevealStartCommand(loader, result, args):
    if loader.isRevealServerLoaded():
        loader.startServer()
    else:
        result.SetError("Reveal Server is not loaded. Use `reveal load` command first.")

def HandleRevealStopCommand(loader, result, args):
    if loader.isRevealServerLoaded():
        loader.stopServer()
    else:
        result.AppendWarning("Reveal Server is not loaded, there's nothing to stop.")
        
def HandleRevealStatusCommand(loader, result, args):
    if loader.isRevealServerLoaded():
        version = loader.getRevealServerProtocolVersion()
        if loader.isRevealServerRunning():
            result.AppendMessage("Reveal Server (protocol version {0}) is loaded and running.".format(version))
        else:
            result.AppendMessage("Reveal Server (protocol version {0}) is loaded, but not running.".format(version))
    else:
        result.AppendMessage("Reveal Server is not loaded.")

# Reveal Loader
class RevealLoader(object):
    def __init__(self, frame):
        self.frame = frame
        self.process = frame.thread.process

    def localRevealServerBinaryPath(self, target, result):
        variants = self.getVariantsInFrameworkAtPath(localFrameworkPath)
        variant = self.getFrameworkVariantForTargetProcess(variants, target, result)

        if variant is not None:
            return os.path.join(localFrameworkPath, variant, "RevealServer.framework/RevealServer")
        else:
            return None

    def remoteRevealServerBinaryPath(self, target, result):
        bundledFrameworkPath = self.remoteFrameworkPath()

        if bundledFrameworkPath is not None and bundledFrameworkPath != "nil":
            variants = self.remoteFrameworkVariants(bundledFrameworkPath)
            variant = self.getFrameworkVariantForTargetProcess(variants, target, result)

            if variant is not None:
                return os.path.join(bundledFrameworkPath, variant, "RevealServer.framework/RevealServer")
            else:
                return None
        else:
            return None
        
    def remoteFileExists(self, remoteFilePath):
        expression = "(BOOL)[[objc_getClass(\"NSFileManager\") defaultManager] fileExistsAtPath:@\"{0}\"] != NO".format(remoteFilePath)
        return self.frame.EvaluateExpression(expression, GetCommonExpressionOptions()).value == "true"

    def remoteFrameworkPath(self):
        expression = "(id)[[objc_getClass(\"NSBundle\") mainBundle] pathForResource:@\"RevealServer\" ofType:@\"xcframework\" inDirectory: @\"Frameworks\"]"
        return self.frame.EvaluateExpression(expression, GetCommonExpressionOptions()).GetObjectDescription()

    def remoteFrameworkVariants(self, path):
        expression = "(NSArray *)[[objc_getClass(\"NSFileManager\") defaultManager] contentsOfDirectoryAtURL:(NSURL *)[[NSURL alloc] initFileURLWithPath:@\"{0}\"] includingPropertiesForKeys: nil options: 0 error: nil]".format(path)
        response = self.frame.EvaluateExpression(expression, GetCommonExpressionOptions()).GetObjectDescription()

        if response is not None and response != "nil":
            return re.findall( r'file:\/\/\/.*\/RevealServer\.xcframework\/([^\/]+)\/,?\n', response, re.M | re.I)
        else:
            return []

    def getVariantsInFrameworkAtPath(self, path):
        return [ f for f in os.listdir(path) if os.path.isdir(os.path.join(path, f)) ]

    def getFrameworkVariantForTargetProcess(self, variants, target, result):
        targetVariantTuple = TargetTripleToVariantTuple(target, result)

        regex = re.compile(r'([^-]+)-([^-]+)(-.*)?', re.M | re.I)

        for variant in variants:
            results = regex.search(variant)

            if results:
                variantPlatform = results.group(1) # ie. ios, tvos, etc
                variantArchitectures = results.group(2) # ie. armv7_arm64_arm64e, x86_64, etc
                variantTarget = str(results.group(3) or '')[1:] # ie. maccatalyst, simulator, or nothing for actual devices

                if targetVariantTuple[0] == variantPlatform and targetVariantTuple[1] in variantArchitectures and targetVariantTuple[2] == variantTarget:
                    return variant
                else:
                    continue

        result.SetError("Could not locate a Reveal Server framework variant for the current target process of type '{0}'. \n\nDid find the following directories within RevealServer.xcframework: \n - {1}.".format(target.triple, '\n - '.join(variants)))

        return None

    def isRevealServerLoaded(self):
        expression = "(void*)dlsym((void*)-2, \"OBJC_CLASS_$_IBARevealLoader\")"
        value = self.frame.EvaluateExpression(expression, GetCommonExpressionOptions()).value

        if value is None:
            return False
        else:
            # Attempt to convert the returned hexadecimal base-16 value to an int
            try:
                pointerValue = int(str(value), 16)
                return pointerValue != 0
            except:
                return False
    
    def getRevealServerProtocolVersion(self):
        expression = "(int)[objc_getClass(\"IBARevealProtocol\") protocolVersion]"
        return self.frame.EvaluateExpression(expression, GetCommonExpressionOptions()).signed

    def injectServer(self, path, result):
        # Check that Reveal Server is already loaded first
        if self.isRevealServerLoaded():
            result.AppendWarning("not loading Reveal Server, it's already loaded.")
            return
        
        # Load Reveal Server image from the specified path
        result.AppendMessage("Loading Reveal Server from {0}...".format(path))
        error = lldb.SBError()
        self.process.LoadImage(lldb.SBFileSpec(path), error)
        
        if error.fail:
            result.SetError(error)
        else:
            result.AppendMessage("Reveal Server was loaded successfully.")

    def isRevealServerRunning(self):
        expression = "(BOOL)[[[objc_getClass(\"IBARevealServer\") sharedServer] httpServer] isRunning] != NO"
        return self.frame.EvaluateExpression(expression, GetCommonExpressionOptions()).value == "true"

    def startServer(self):
        self.frame.EvaluateExpression("(void)[objc_getClass(\"IBARevealLoader\") startServer]", GetCommonExpressionOptions())

    def stopServer(self):
        self.frame.EvaluateExpression("(void)[objc_getClass(\"IBARevealLoader\") stopServer]", GetCommonExpressionOptions())

# Target info utilities
def TargetIsCompatible(target):
    architecture = target.triple
    return "apple-ios" in architecture or "apple-tvos" in architecture or "apple-watchos" in architecture

def TargetIsSimulator(target):
    platformName = target.platform.GetName()
    return platformName.endswith("simulator")

def TargetTripleToVariantTuple(target, result):
    # Extract the architecture, platform and target from the provided target's triple:
    searchResults = re.search( r'([^-]+)-[^-]+-([a-z]+)(-.*)?', target.triple, re.M | re.I)

    if searchResults is None:
        result.SetError("Could not determine enough information about the current process to determine which Reveal Server framework variant to use")
        return ""

    foundArchitecture = searchResults.group(1) # ie. x86_64, armv7, etc
    foundPlatform = searchResults.group(2) # ie. ios, tvos
    foundTarget = str(searchResults.group(3) or '')[1:] # ie. macabi, or nothing for all other targets

    # We don't actually ship an x86_64h slice, so redact the "h"
    if foundArchitecture == "x86_64h":
        foundArchitecture = "x86_64"

    # There is a mismatch between what LLDB refers to Mac Catalyst apps as, and how Xcode names XCFramework targets. Workaround this difference:
    if foundTarget == "macabi":
        foundTarget = "maccatalyst"

    # Append the "simulator" suffix if the target is a Simulator
    if foundTarget == "" and target.platform.GetName().endswith("simulator"):
        foundTarget = "simulator"

    return (foundPlatform, foundArchitecture, foundTarget)

# Process info utilities
def ProcessMainThreadContainsFrameWithName(process, frameName):
    mainThreadFrames = process.GetThreadAtIndex(0).frames
    for frame in mainThreadFrames:
        if frame.name == frameName:
            return True

    return False

def ProcessFirstFrameNameContains(process, frameName):
    mainThreadFrames = process.GetThreadAtIndex(0).frames
    return mainThreadFrames[0].name == frameName

# Debugger utilities
def GetCommonExpressionOptions():
    options = lldb.SBExpressionOptions()
    options.SetLanguage(lldb.eLanguageTypeObjC)
    options.SetSuppressPersistentResult(True)
    return options

def IsLegacyBreakpointAtProcessFirstFrame(process):
    return ProcessFirstFrameNameContains(process, "UIApplicationMain") or ProcessFirstFrameNameContains(process, "-[UIApplication _run]")

def SetupLegacyLoadBreakpoints(target, autostart, result):
    if autostart:
        callback_name = load_reveal_autostart_callback.__qualname__
    else:
        callback_name = load_reveal_callback.__qualname__

    # Import the callback so that LLDB can see it:
    lldb.debugger.HandleCommand(
        "script from %s import %s" % (__name__, callback_name)
    )

    symbol_names = [
        '-[UIViewController viewDidLoad]',
        'UIViewController.viewDidLoad()'
    ]

    breakpoint_registration_count = 0
    
    for symbol_name in symbol_names:
        breakpoint = target.BreakpointCreateByName(symbol_name)
        breakpoint.SetScriptCallbackFunction(callback_name)    
        breakpoint.SetAutoContinue(True)
        breakpoint.SetOneShot(True)  
        breakpoint_registration_count += breakpoint.GetNumResolvedLocations() 

    if breakpoint_registration_count == 0:
        result.SetError("Could not find any breakpoints to register for the Reveal Server framework. Tried the following symbols: {0}".format(", ".join(symbol_names)))
    
def load_reveal_callback(frame, bp_loc, internal_dict):
    lldb.debugger.HandleCommand("reveal load")

def load_reveal_autostart_callback(frame, bp_loc, internal_dict):
    lldb.debugger.HandleCommand("reveal load --autostart")
